@@ -1,5 +1,5 @@
 /*
-** © 2009-2017 by Kornel Lesiński.
+** © 2009-2018 by Kornel Lesiński.
 ** © 1989, 1991 by Jef Poskanzer.
 ** © 1997, 2000, 2002 by Greg Roelofs; based on an idea by Stefan Schneider.
 **
@@ -21,7 +21,9 @@
 
 #ifdef _OPENMP
 #include <omp.h>
+#define LIQ_TEMP_ROW_WIDTH(img_width) (((img_width) | 15) + 1) /* keep alignment & leave space between rows to avoid cache line contention */
 #else
+#define LIQ_TEMP_ROW_WIDTH(img_width) (img_width)
 #define omp_get_max_threads() 1
 #define omp_get_thread_num() 0
 #endif
@@ -56,7 +58,8 @@ struct liq_attr {
     unsigned int max_colors, max_histogram_entries;
     unsigned int min_posterization_output /* user setting */, min_posterization_input /* speed setting */;
     unsigned int kmeans_iterations, feedback_loop_trials;
-    bool last_index_transparent, use_contrast_maps, use_dither_map;
+    bool last_index_transparent, use_contrast_maps;
+    unsigned char use_dither_map;
     unsigned char speed;
 
     unsigned char progress_stage1, progress_stage2, progress_stage3;
@@ -103,7 +106,8 @@ typedef struct liq_remapping_result {
     liq_palette int_palette;
     double gamma, palette_error;
     float dither_level;
-    bool use_dither_map; unsigned char progress_stage1;
+    unsigned char use_dither_map;
+    unsigned char progress_stage1;
 } liq_remapping_result;
 
 struct liq_result {
@@ -120,7 +124,7 @@ struct liq_result {
     float dither_level;
     double gamma, palette_error;
     int min_posterization_output;
-    bool use_dither_map;
+    unsigned char use_dither_map;
 };
 
 struct liq_histogram {
@@ -140,6 +144,7 @@ static void modify_alpha(liq_image *input_image, rgba_pixel *const row_pixels) L
 static void contrast_maps(liq_image *image) LIQ_NONNULL;
 static liq_error finalize_histogram(liq_histogram *input_hist, liq_attr *options, histogram **hist_output) LIQ_NONNULL;
 static const rgba_pixel *liq_image_get_row_rgba(liq_image *input_image, unsigned int row) LIQ_NONNULL;
+static bool liq_image_get_row_f_init(liq_image *img) LIQ_NONNULL;
 static const f_pixel *liq_image_get_row_f(liq_image *input_image, unsigned int row) LIQ_NONNULL;
 static void liq_remapping_result_destroy(liq_remapping_result *result) LIQ_NONNULL;
 static liq_error pngquant_quantize(histogram *hist, const liq_attr *options, const int fixed_colors_count, const f_pixel fixed_colors[], const double gamma, bool fixed_result_colors, liq_result **) LIQ_NONNULL;
@@ -153,7 +158,7 @@ LIQ_NONNULL static void liq_verbose_printf(const liq_attr *context, const char *
         int required_space = vsnprintf(NULL, 0, fmt, va)+1; // +\0
         va_end(va);
 
-        char buf[required_space];
+        LIQ_ARRAY(char, buf, required_space);
         va_start(va, fmt);
         vsnprintf(buf, required_space, fmt, va);
         va_end(va);
@@ -189,8 +194,13 @@ LIQ_NONNULL static bool liq_remap_progress(const liq_remapping_result *quant, co
 #if USE_SSE
 inline static bool is_sse_available()
 {
-#if (defined(__x86_64__) || defined(__amd64))
+#if (defined(__x86_64__) || defined(__amd64) || defined(_WIN64))
     return true;
+#elif _MSC_VER
+    int info[4];
+    __cpuid(info, 1);
+    /* bool is implemented as a built-in type of size 1 in MSVC */
+    return info[3] & (1<<26) ? true : false;
 #else
     int a,b,c,d;
         cpuid(1, a, b, c, d);
@@ -327,7 +337,8 @@ LIQ_EXPORT LIQ_NONNULL liq_error liq_set_speed(liq_attr* attr, int speed)
     if (!CHECK_STRUCT_TYPE(attr, liq_attr)) return LIQ_INVALID_POINTER;
     if (speed < 1 || speed > 10) return LIQ_VALUE_OUT_OF_RANGE;
 
-    unsigned int iterations = MAX(8-speed, 0); iterations += iterations * iterations/2;
+    unsigned int iterations = MAX(8-speed, 0);
+    iterations += iterations * iterations/2;
     attr->kmeans_iterations = iterations;
     attr->kmeans_iteration_limit = 1.0/(double)(1<<(23-speed));
     attr->feedback_loop_trials = MAX(56-9*speed, 0);
@@ -335,11 +346,16 @@ LIQ_EXPORT LIQ_NONNULL liq_error liq_set_speed(liq_attr* attr, int speed)
     attr->max_histogram_entries = (1<<17) + (1<<18)*(10-speed);
     attr->min_posterization_input = (speed >= 8) ? 1 : 0;
     attr->use_dither_map = (speed <= (omp_get_max_threads() > 1 ? 7 : 5)); // parallelized dither map might speed up floyd remapping
+    if (attr->use_dither_map && speed < 3) {
+        attr->use_dither_map = 2; // always
+    }
     attr->use_contrast_maps = (speed <= 7) || attr->use_dither_map;
     attr->speed = speed;
 
     attr->progress_stage1 = attr->use_contrast_maps ? 20 : 8;
-    if (attr->feedback_loop_trials < 2) attr->progress_stage1 += 30;
+    if (attr->feedback_loop_trials < 2) {
+        attr->progress_stage1 += 30;
+    }
     attr->progress_stage3 = 50 / (1+speed);
     attr->progress_stage2 = 100 - attr->progress_stage1 - attr->progress_stage3;
     return LIQ_OK;
@@ -499,7 +515,7 @@ LIQ_EXPORT liq_attr* liq_attr_create_with_allocator(void* (*custom_malloc)(size_
         .target_mse = 0,
         .max_mse = MAX_DIFF,
     };
-    liq_set_speed(attr, 3);
+    liq_set_speed(attr, 4);
     return attr;
 }
 
@@ -519,7 +535,7 @@ LIQ_EXPORT LIQ_NONNULL liq_error liq_image_add_fixed_color(liq_image *img, liq_c
     return LIQ_OK;
 }
 
-LIQ_NONNULL liq_error liq_histogram_add_fixed_color(liq_histogram *hist, f_pixel color)
+LIQ_NONNULL static liq_error liq_histogram_add_fixed_color_f(liq_histogram *hist, f_pixel color)
 {
     if (hist->fixed_colors_count > 255) return LIQ_UNSUPPORTED;
 
@@ -527,9 +543,24 @@ LIQ_NONNULL liq_error liq_histogram_add_fixed_color(liq_histogram *hist, f_pixel
     return LIQ_OK;
 }
 
+LIQ_EXPORT LIQ_NONNULL liq_error liq_histogram_add_fixed_color(liq_histogram *hist, liq_color color, double gamma)
+{
+    if (!CHECK_STRUCT_TYPE(hist, liq_histogram)) return LIQ_INVALID_POINTER;
+
+    float gamma_lut[256];
+    to_f_set_gamma(gamma_lut, gamma ? gamma : 0.45455);
+    const f_pixel px = rgba_to_f(gamma_lut, (rgba_pixel){
+        .r = color.r,
+        .g = color.g,
+        .b = color.b,
+        .a = color.a,
+    });
+    return liq_histogram_add_fixed_color_f(hist, px);
+}
+
 LIQ_NONNULL static bool liq_image_use_low_memory(liq_image *img)
 {
-    img->temp_f_row = img->malloc(sizeof(img->f_pixels[0]) * img->width * omp_get_max_threads());
+    img->temp_f_row = img->malloc(sizeof(img->f_pixels[0]) * LIQ_TEMP_ROW_WIDTH(img->width) * omp_get_max_threads());
     return img->temp_f_row != NULL;
 }
 
@@ -565,7 +596,7 @@ static liq_image *liq_image_create_internal(const liq_attr *attr, rgba_pixel* ro
     };
 
     if (!rows || attr->min_opaque_val < 1.f) {
-        img->temp_row = attr->malloc(sizeof(img->temp_row[0]) * width * omp_get_max_threads());
+        img->temp_row = attr->malloc(sizeof(img->temp_row[0]) * LIQ_TEMP_ROW_WIDTH(width) * omp_get_max_threads());
         if (!img->temp_row) return NULL;
     }
 
@@ -760,7 +791,7 @@ LIQ_NONNULL static const rgba_pixel *liq_image_get_row_rgba(liq_image *img, unsi
     }
 
     assert(img->temp_row);
-    rgba_pixel *temp_row = img->temp_row + img->width * omp_get_thread_num();
+    rgba_pixel *temp_row = img->temp_row + LIQ_TEMP_ROW_WIDTH(img->width) * omp_get_thread_num();
     if (img->rows) {
         memcpy(temp_row, img->rows[row], img->width * sizeof(temp_row[0]));
     } else {
@@ -783,31 +814,40 @@ LIQ_NONNULL static void convert_row_to_f(liq_image *img, f_pixel *row_f_pixels,
     }
 }
 
+LIQ_NONNULL static bool liq_image_get_row_f_init(liq_image *img)
+{
+    assert(omp_get_thread_num() == 0);
+    if (img->f_pixels) {
+        return true;
+    }
+    if (!liq_image_should_use_low_memory(img, false)) {
+        img->f_pixels = img->malloc(sizeof(img->f_pixels[0]) * img->width * img->height);
+    }
+    if (!img->f_pixels) {
+        return liq_image_use_low_memory(img);
+    }
+
+    if (!liq_image_has_rgba_pixels(img)) {
+        return false;
+    }
+
+    float gamma_lut[256];
+    to_f_set_gamma(gamma_lut, img->gamma);
+    for(unsigned int i=0; i < img->height; i++) {
+        convert_row_to_f(img, &img->f_pixels[i*img->width], i, gamma_lut);
+    }
+    return true;
+}
+
 LIQ_NONNULL static const f_pixel *liq_image_get_row_f(liq_image *img, unsigned int row)
 {
     if (!img->f_pixels) {
-        if (img->temp_f_row) {
-            float gamma_lut[256];
-            to_f_set_gamma(gamma_lut, img->gamma);
-            f_pixel *row_for_thread = img->temp_f_row + img->width * omp_get_thread_num();
-            convert_row_to_f(img, row_for_thread, row, gamma_lut);
-            return row_for_thread;
-        }
-
-        assert(omp_get_thread_num() == 0);
-        if (!liq_image_should_use_low_memory(img, false)) {
-            img->f_pixels = img->malloc(sizeof(img->f_pixels[0]) * img->width * img->height);
-        }
-        if (!img->f_pixels) {
-            if (!liq_image_use_low_memory(img)) return NULL;
-            return liq_image_get_row_f(img, row);
-        }
-
+        assert(img->temp_f_row); // init should have done that
         float gamma_lut[256];
         to_f_set_gamma(gamma_lut, img->gamma);
-        for(unsigned int i=0; i < img->height; i++) {
-            convert_row_to_f(img, &img->f_pixels[i*img->width], i, gamma_lut);
-        }
+        f_pixel *row_for_thread = img->temp_f_row + LIQ_TEMP_ROW_WIDTH(img->width) * omp_get_thread_num();
+        convert_row_to_f(img, row_for_thread, row, gamma_lut);
+        return row_for_thread;
     }
     return img->f_pixels + img->width * row;
 }
@@ -1219,7 +1259,10 @@ LIQ_NONNULL static float remap_to_palette(liq_image *const input_image, unsigned
     const unsigned int cols = input_image->width;
     double remapping_error=0;
 
-    if (!liq_image_get_row_f(input_image, 0)) { // trigger lazy conversion
+    if (!liq_image_get_row_f_init(input_image)) {
+        return -1;
+    }
+    if (input_image->background && !liq_image_get_row_f_init(input_image->background)) {
         return -1;
     }
 
@@ -1230,7 +1273,7 @@ LIQ_NONNULL static float remap_to_palette(liq_image *const input_image, unsigned
 
 
     const unsigned int max_threads = omp_get_max_threads();
-    kmeans_state average_color[(KMEANS_CACHE_LINE_GAP+map->colors) * max_threads];
+    LIQ_ARRAY(kmeans_state, average_color, (KMEANS_CACHE_LINE_GAP+map->colors) * max_threads);
     kmeans_init(map, max_threads, average_color);
 
     #pragma omp parallel for if (rows*cols > 3000) \
@@ -1292,14 +1335,14 @@ inline static f_pixel get_dithered_pixel(const float dither_level, const float m
      } else if (dither_error < 2.f/256.f/256.f) {
         // don't dither areas that don't have noticeable error — makes file smaller
         return px;
-     }
+    }
 
-     return (f_pixel){
-         .r=px.r + sr * ratio,
-         .g=px.g + sg * ratio,
-         .b=px.b + sb * ratio,
-         .a=a,
-     };
+    return (f_pixel) {
+        .r=px.r + sr * ratio,
+        .g=px.g + sg * ratio,
+        .b=px.b + sb * ratio,
+        .a=a,
+    };
 }
 
 /**
@@ -1315,6 +1358,13 @@ LIQ_NONNULL static bool remap_to_palette_floyd(liq_image *input_image, unsigned
     const colormap *map = quant->palette;
     const colormap_item *acolormap = map->palette;
 
+    if (!liq_image_get_row_f_init(input_image)) {
+        return false;
+    }
+    if (input_image->background && !liq_image_get_row_f_init(input_image->background)) {
+        return false;
+    }
+
     /* Initialize Floyd-Steinberg error vectors. */
     const size_t errwidth = cols+2;
     f_pixel *restrict thiserr = input_image->malloc(errwidth * sizeof(thiserr[0]) * 2); // +2 saves from checking out of bounds access
@@ -1531,19 +1581,21 @@ LIQ_EXPORT LIQ_NONNULL liq_error liq_histogram_add_image(liq_histogram *input_hi
     input_hist->gamma = input_image->gamma;
 
     for(int i = 0; i < input_image->fixed_colors_count; i++) {
-        liq_error res = liq_histogram_add_fixed_color(input_hist, input_image->fixed_colors[i]);
+        liq_error res = liq_histogram_add_fixed_color_f(input_hist, input_image->fixed_colors[i]);
         if (res != LIQ_OK) {
             return res;
         }
     }
 
-   /*
-    ** Step 2: attempt to make a histogram of the colors, unclustered.
-    ** If at first we don't succeed, increase ignorebits to increase color
-    ** coherence and try again.
-    */
+    /*
+     ** Step 2: attempt to make a histogram of the colors, unclustered.
+     ** If at first we don't succeed, increase ignorebits to increase color
+     ** coherence and try again.
+     */
 
-    if (liq_progress(options, options->progress_stage1 * 0.4f)) return LIQ_ABORTED;
+    if (liq_progress(options, options->progress_stage1 * 0.4f)) {
+        return LIQ_ABORTED;
+    }
 
     const bool all_rows_at_once = liq_image_can_use_rgba_rows(input_image);
 
@@ -1655,7 +1707,7 @@ LIQ_NONNULL static void contrast_maps(liq_image *image)
 
     unsigned char *restrict tmp = image->malloc(cols*rows);
 
-    if (!noise || !edges || !tmp) {
+    if (!noise || !edges || !tmp || !liq_image_get_row_f_init(image)) {
         image->free(noise);
         image->free(edges);
         image->free(tmp);
@@ -1822,6 +1874,10 @@ static colormap *find_best_palette(histogram *hist, const liq_attr *options, con
     // at this point actual gamma is not set, so very conservative posterization estimate is used
     const double target_mse = MIN(max_mse, MAX(options->target_mse, pow((1<<options->min_posterization_output)/1024.0, 2)));
     int feedback_loop_trials = options->feedback_loop_trials;
+    if (hist->size > 5000) {feedback_loop_trials = (feedback_loop_trials*3 + 3)/4;}
+    if (hist->size > 25000) {feedback_loop_trials = (feedback_loop_trials*3 + 3)/4;}
+    if (hist->size > 50000) {feedback_loop_trials = (feedback_loop_trials*3 + 3)/4;}
+    if (hist->size > 100000) {feedback_loop_trials = (feedback_loop_trials*3 + 3)/4;}
     colormap *acolormap = NULL;
     double least_error = MAX_DIFF;
     double target_mse_overshoot = feedback_loop_trials>0 ? 1.05 : 1.0;
@@ -1831,7 +1887,7 @@ static colormap *find_best_palette(histogram *hist, const liq_attr *options, con
         colormap *newmap;
         if (hist->size && fixed_colors_count < max_colors) {
             newmap = mediancut(hist, max_colors-fixed_colors_count, target_mse * target_mse_overshoot, MAX(MAX(45.0/65536.0, target_mse), least_error)*1.2,
-            options->malloc, options->free);
+                               options->malloc, options->free);
         } else {
             feedback_loop_trials = 0;
             newmap = NULL;
@@ -1928,7 +1984,7 @@ LIQ_NONNULL static liq_error pngquant_quantize(histogram *hist, const liq_attr *
         }
 
         // K-Means iteration approaches local minimum for the palette
-        const double iteration_limit = options->kmeans_iteration_limit;
+        double iteration_limit = options->kmeans_iteration_limit;
         unsigned int iterations = options->kmeans_iterations;
 
         if (!iterations && palette_error < 0 && max_mse < MAX_DIFF) iterations = 1; // otherwise total error is never calculated and MSE limit won't work
@@ -1941,6 +1997,11 @@ LIQ_NONNULL static liq_error pngquant_quantize(histogram *hist, const liq_attr *
                 }
             }
 
+            if (hist->size > 5000) {iterations = (iterations*3 + 3)/4;}
+            if (hist->size > 25000) {iterations = (iterations*3 + 3)/4;}
+            if (hist->size > 50000) {iterations = (iterations*3 + 3)/4;}
+            if (hist->size > 100000) {iterations = (iterations*3 + 3)/4; iteration_limit *= 2;}
+
             verbose_print(options, "  moving colormap towards local minimum");
 
             double previous_palette_error = MAX_DIFF;
@@ -2022,7 +2083,7 @@ LIQ_EXPORT LIQ_NONNULL liq_error liq_write_remapped_image(liq_result *result, li
         return LIQ_BUFFER_TOO_SMALL;
     }
 
-    unsigned char *rows[input_image->height];
+    LIQ_ARRAY(unsigned char *, rows, input_image->height);
     unsigned char *buffer_bytes = buffer;
     for(unsigned int i=0; i < input_image->height; i++) {
         rows[i] = &buffer_bytes[input_image->width * i];
@@ -2062,7 +2123,9 @@ LIQ_EXPORT LIQ_NONNULL liq_error liq_write_remapped_image_rows(liq_result *quant
         set_rounded_palette(&result->int_palette, result->palette, result->gamma, quant->min_posterization_output);
         remapping_error = remap_to_palette(input_image, row_pointers, result->palette);
     } else {
-        const bool generate_dither_map = result->use_dither_map && (input_image->edges && !input_image->dither_map);
+        const bool is_image_huge = (input_image->width * input_image->height) > 2000 * 2000;
+        const bool allow_dither_map = result->use_dither_map == 2 || (!is_image_huge && result->use_dither_map);
+        const bool generate_dither_map = allow_dither_map && (input_image->edges && !input_image->dither_map);
         if (generate_dither_map) {
             // If dithering (with dither map) is required, this image is used to find areas that require dithering
             remapping_error = remap_to_palette(input_image, row_pointers, result->palette);
